## DESCRIPTION: R scripts and functions that can be used to reproduce the
## results in the paper
## Kosmidis (2013). Improved estimation in cumulative link models.
## AUTHOR: Dr Ioannis Kosmidis
## DATE: Jan 2013
## EMAIL: i.kosmidis@ucl.ac.uk
## WEB: http://www.ucl.ac.uk/~ucakiko


## The current script should be run after running script Section10.R
## for each of the values 1, 2, 3, 4, 5, 6, 7, 8, 9, 10 for whichseed.

rm(list = ls())
source("cImages.R")
library(MASS)
library(ordinal)
source("bpolr.R")
index <- c(1, 2, 3, 4, 5, 6, 7, 8, 9, 10)
wd <- paste0(getwd(), "/images")


imageNames <- paste("admit_", index, ".RData", sep = "")
cImages(imageNames, c("betasML", "betasBR", "betasBC",
                      "stderrsML", "stderrsBR", "stderrsBC",
                      "boundaryEstimatesML",
                      "boundaryEstimatesBR",
                      "boundaryEstimatesBC",
                      "inadmissibleML",
                      "inadmissibleBR",
                      "inadmissibleBC",
                      "simulatedData",
                      "nsimu"),
        working.directory = wd)

cImages(imageNames[1], c("truth", "nam", "fmML"), working.directory = wd)

cbindObjects("nsimu", ".image", 1:length(index), delete = TRUE)
cbindObjects("betasML", ".image", 1:length(index), delete = TRUE)
cbindObjects("betasBR", ".image", 1:length(index), delete = TRUE)
cbindObjects("betasBC", ".image", 1:length(index), delete = TRUE)
cbindObjects("stderrsML", ".image", 1:length(index), delete = TRUE)
cbindObjects("stderrsBR", ".image", 1:length(index), delete = TRUE)
cbindObjects("stderrsBC", ".image", 1:length(index), delete = TRUE)
cObjects("boundaryEstimatesML", ".image", 1:length(index), delete = TRUE)
cObjects("boundaryEstimatesBR", ".image", 1:length(index), delete = TRUE)
cObjects("boundaryEstimatesBC", ".image", 1:length(index), delete = TRUE)
#cbindObjects("simulatedData", ".image", 1:length(index), delete = TRUE)
cObjects("truth", ".image", 1, delete = TRUE)
fmML <- fmML.image1

## Coverage probabilities
cover <- function(truth, coefs, ses, quantile = qnorm(0.975)) {
  low <- coefs - quantile*ses
  upp <- coefs + quantile*ses
  mean((low < truth) & (upp > truth), na.rm = TRUE)
}

cover1 <- function(truth, coefs, ses, quantile = qnorm(0.975)) {
  low <- coefs - quantile*ses
  upp <- coefs + quantile*ses
  var((low < truth) & (upp > truth), na.rm = TRUE)*(length(coefs) - 1)/length(coefs)
}


coverageML <- sapply(1:length(truth), function(k) {
  cover(truth[k], betasML[k, ], stderrsML[k, ], qnorm(0.975))
})

coverageBR <- sapply(1:length(truth), function(k) {
  cover(truth[k], betasBR[k, ], stderrsBR[k, ], qnorm(0.975))
})

coverageBC <- sapply(1:length(truth), function(k) {
  cover(truth[k], betasBC[k, ], stderrsBC[k, ], qnorm(0.975))
})

coverage <- cbind(coverageML, coverageBC, coverageBR)
dimnames(coverage) <- list(names(truth), c("ML", "BC", "BR"))


## Remove infinite estimates
infiniteEstimates <- apply(betasML, 2, function(x) any(abs(x) > 20)) &
  apply(stderrsML, 2, function(x) any(abs(x) > 100))
betasMLnoInf <- betasML[, !infiniteEstimates]
betasBCnoInf <- betasBC[, !infiniteEstimates]

biasML <- apply(betasMLnoInf, 1, mean) - truth
biasBR <- apply(betasBR, 1, mean) - truth
biasBC <- apply(betasBCnoInf, 1, mean) - truth

bias <- cbind(biasML, biasBC, biasBR)
dimnames(bias) <- list(names(truth), c("ML", "BC", "BR"))

varianceML <- apply(betasMLnoInf, 1, var)*(ncol(betasMLnoInf) - 1)/ncol(betasMLnoInf)
varianceBR <- apply(betasBR, 1, var)*(ncol(betasBR) - 1)/ncol(betasBR)
varianceBC <- apply(betasBCnoInf, 1, var)*(ncol(betasBCnoInf) - 1)/ncol(betasBCnoInf)

variance <- cbind(varianceML, varianceBC, varianceBR)
dimnames(variance) <- list(names(truth), c("ML", "BC", "BR"))

mseML <- apply((betasMLnoInf - truth)^2, 1, mean)
mseBR <- apply((betasBR - truth)^2, 1, mean)
mseBC <- apply((betasBCnoInf - truth)^2, 1, mean)

mse <- cbind(mseML, mseBC, mseBR)
dimnames(mse) <- list(names(truth), c("ML", "BC", "BR"))


## The mean squared error is larger by this percent from its
## minimum value --- the variance --- due to bias
msePerc <- 100*bias^2/variance



## Monte Carlo errors for bias estimator
MCerrorBiasML <- sqrt(apply(betasMLnoInf, 1, var)/ncol(betasMLnoInf))
MCerrorBiasBR <- sqrt(apply(betasBR, 1, var)/ncol(betasBR))
MCerrorBiasBC <- sqrt(apply(betasBCnoInf, 1, var)/ncol(betasBCnoInf))

MCerrorBias <- cbind(MCerrorBiasML, MCerrorBiasBC, MCerrorBiasBR)
dimnames(MCerrorBias) <- list(names(truth), c("ML", "BC", "BR"))

## Monte Carlo errors for mean squared error estimator
MCerrorMSEML <- sqrt(apply((betasMLnoInf - truth)^2, 1, var)/ncol(betasMLnoInf))
MCerrorMSEBR <- sqrt(apply((betasBR - truth)^2, 1, var)/ncol(betasBR))
MCerrorMSEBC <- sqrt(apply((betasBCnoInf - truth)^2, 1, var)/ncol(betasBCnoInf))

MCerrorMSE <- cbind(MCerrorMSEML, MCerrorMSEBC, MCerrorMSEBR)
dimnames(MCerrorMSE) <- list(names(truth), c("ML", "BC", "BR"))

## Monte Carlo errors for coverage
MCerrorCoverage <- sqrt(coverage*(1-coverage)/ncol(betasBR))
dimnames(MCerrorCoverage) <- list(names(truth), c("ML", "BC", "BR"))

res <- cbind(c(bias[,1], bias[,2], bias[,3]),
             c(mse[,1], mse[,2], mse[,3]),
             c(coverage[,1], coverage[,2], coverage[,3]),
             c(msePerc[,1], msePerc[,2], msePerc[,3]))

