## DESCRIPTION: R scripts and functions that can be used to reproduce the
## results in the paper
## Kosmidis (2013). Improved estimation in cumulative link models.
## AUTHOR: Dr Ioannis Kosmidis
## DATE: Jan 2013
## EMAIL: i.kosmidis@ucl.ac.uk
## WEB: http://www.ucl.ac.uk/~ucakiko


rm(list = ls())
source("bpolr.R")
require(combinat)
require(MASS)

## Set up for one of the complete enumerations in the paper (1c, 2c,
## 3c, 1s, 2s). A type specification other than the latter 5 will
## default to the setting shown under comment 123 below. There you may
## set-up your custom complete enumeration exercise.
type <- "2s"

## Types 1c, 2c, 3c should all be used to perform the complete
## enumerations that correspond to the results in Section 8 of the
## paper. The script will save the results under the directory images,
## which should be available under the current working directory.
## Then completeEnumeration_plots.R should be run to produce the
## corresponding plots in the paper.

## Types 1s, 2s should both be used to perform the complete enumerations that
## are required for Section 9 of the paper. The script will save the
## results under the directory images, which should be available under
## the current working directory.
## Then shrinkage_plots.R should be run to produce the
## shrinkage plots in the paper.

if (type == "1c") {
  ncat <- 4
  tots <- c(3, 3)
  Y <- factor(rep(1:ncat, length(tots)))
  x <- rep(c(-0.5, 0.5), each = ncat)
  link <- "logit"
}
if (type == "2c") {
  ncat <- 4
  tots <- c(5, 5)
  Y <- factor(rep(1:ncat, length(tots)))
  x <- rep(c(-0.5, 0.5), each = ncat)
  link <- "logit"
}
if (type == "3c") {
  ncat <- 4
  tots <- c(10, 10)
  Y <- factor(rep(1:ncat, length(tots)))
  x <- rep(c(-0.5, 0.5), each = ncat)
  link <- "logit"
}
if (type == "1s") {
  ncat <- 6
  tots <- c(3, 3)
  Y <- factor(rep(1:ncat, length(tots)))
  x <- rep(c(-0.5, 0.5), each = ncat)
  link <- "logit"
}
if (type == "2s") {
  ncat <- 6
  tots <- c(3, 3)
  Y <- factor(rep(1:ncat, length(tots)))
  x <- rep(c(-0.5, 0.5), each = ncat)
  link <- "cloglog"
}
## 123: Set up custom complete enumeration settings
if (!(type %in% c(paste0(1:3, "c"), paste0(1:2, "s")))) {
    warning("A non-standard type of complete enumeration has been specified")
    ## Number of categories
    ncat <- 2
    ## Number of multinomial vectors
    tots <- c(2, 2)
    ## Ordinal responses
    Y <- factor(rep(1:ncat, length(tots)))
    ## Covariate settings. Should be the same length as Y
    x <- rep(c(-0.5, 0.5), each = ncat)
    ## Link function
    link <- "logit"
}

################################################
## Some necessary functions
################################################
allTables <- function(ncat = 2, rowTotals = c(1, 1),
                      retList = FALSE, dimOnly = TRUE) {
  nxlev <- length(rowTotals)
  if (dimOnly) {
    nmats <- prod(choose(rowTotals + ncat - 1, rowTotals))
    if (retList) {
      cat("The resultant object will be a list of", nmats, ",", paste(ncat, "x", nxlev), "matrices\n")
    }
    else {
      cat("The resultant object will be a matrix of dimension",
          paste(ncat*nxlev, "x", nmats), "\n")
    }
  }
  else {
    require(combinat)
    res <- as.list(rep(NA, nxlev))
    for (i in seq_len(nxlev)) {
      res[[i]] <- xsimplex(ncat, rowTotals[i])
    }
    combine <- function(mat1, mat2) {
      p1 <- ncol(mat1)
      n2 <- nrow(mat2)
      iii <- matrix(1, nrow = n2, ncol = p1)
      out <- NULL
      for (ind in seq_len(ncol(mat2))) {
        out <- cbind(out, rbind(mat1,
                                mat2[, ind] * iii))
      }
      out
    }
    temp <- res[[1]]
    for (i in 1:(nxlev - 1)) {
      temp <- combine(temp, res[[i + 1]])
    }
    makeList <- function(out) {
      ncols <- ncol(out)
      resList <- as.list(rep(NA, ncols))
      for (j in seq_len(ncols)) {
        resList[[j]] <- matrix(out[, j], nrow = ncat, ncol = nxlev)
        dimnames(resList[[j]]) <- list(paste("cat", seq_len(ncat), sep = ""),
                                       paste("x", seq_len(nxlev), sep = ""))
      }
      resList
    }
    if (retList) makeList(temp) else temp
  }
}

empLogist <- function(Y, x, counts) {
  contTable <- xtabs(counts ~ x + Y)
  nlev <- nlevels(Y)
  catTots <- colSums(contTable)
  cumCounts <- apply(contTable, 1, cumsum)
  logits <- apply(cumCounts, 2, function(x) log((x+1/2)/(x[nlev] - x + 1/2)))
  cumCounts1 <- rowSums(cumCounts)
  mat <- diag(nlev) + rbind(cbind(0, diag(nlev - 1)), 0)
  weights <- (cumCounts1*(sum(catTots) - cumCounts1) * mat%*%catTots)[1:(nlev - 1)]
  weights1 <- c(weights/sum(weights))
  if (all(weights == 0)) {
    weights <- weights1 <- rep(1/(nlev - 1), nlev - 1)
  }
  stderr <- sqrt(sum(catTots)^4/(sum(weights) *
                                 prod(cumCounts[nlev, ])))
  list(estimate = -diff(colSums(weights1*logits[1:(nlev - 1), , drop = FALSE])),
       stderr = stderr)
}

## fun is a list with elements c(on what, what to do, result name)
stripObject <- function(obj, keep,
                        fun = list()) {
  whatIsIn <- names(obj)
  newObj <- obj[match(keep, whatIsIn, nomatch = 0)]
  for (i in seq.int(length.out = length(fun))) {
    curFun <- fun[[i]]
    newObj[[curFun[3]]] <- do.call(curFun[2], newObj[curFun[1]])
    newObj[[curFun[1]]] <- NULL
  }
  newObj
}

stdCalc <- function(vcov) {
  if (inherits(vcov, "try-error")) NA else sqrt(vcov[1, 1])
}

identity <- function(beta) {
  beta
}

keep <- c("beta", "vcov",
          "fitted.values",
          "scores", "adjustedScores",
          "boundaryEstimates",
          "inadmissible",
          "convergence",
          "dataReshaped",
          "niter")
fun <- list(c("vcov", "stdCalc", "stderr"),
            c("beta", "identity", "estimate"))

## Generate all data sets
dat <- allTables(ncat = ncat, rowTotals = tots, retList = FALSE,
                 dimOnly = FALSE)


reportEvery <- 5
## Estimation of beta
resEmp <- resML <- resBR <- resBC <- as.list(numeric(ncol(dat)))
for (i in 1:ncol(dat)) {
  tempDat <- data.frame(Y, x, counts = dat[, i])
  ## Empirical logistic transform
  resEmp[[i]] <- empLogist(Y, x, dat[, i])
  ## ML Fit
  resML[[i]] <- stripObject(bpolr(Y ~ x, weights = counts, method = "ML",
                                  data = tempDat, link = link,
                                  epsilon = 1e-10, maxit = 1000),
                          keep, fun)
  ## BC fit
  resBC[[i]] <- stripObject(bpolr(Y ~ x, weights = counts, method = "BC",
                                  data = tempDat, link = link,
                                  epsilon = 1e-10, maxit = 1000),
                          keep, fun)
  resBR[[i]] <- stripObject(bpolr(Y ~ x, weights = counts, method = "BR",
                                  data = tempDat, link = link,
                                  epsilon = 1e-10, maxit = 1000),
                          keep, fun)
  if (i %% reportEvery == 0) {
    cat("===================\n")
    cat("Data set:", i, "out of", ncol(dat), "\n")
    print(xtabs(counts ~ x + Y, data = tempDat))
    cat("Max score:", round(max(abs(resML[[i]]$scores), na.rm = TRUE), 8), "\n")
    cat("Max adjusted score:", round(max(abs(resBR[[i]]$adjustedScores)), 8), "\n")
  }
}

betasML <- simplify2array(lapply(resML, function(x) x$estimate))
betasBC <- simplify2array(lapply(resBC, function(x) x$estimate))
betasBR <- simplify2array(lapply(resBR, function(x) x$estimate))
betasEL <- simplify2array(lapply(resEmp, function(x) x$estimate))

stderrsML <- simplify2array(lapply(resML, function(x) x$stderr))
stderrsBC <- simplify2array(lapply(resBC, function(x) x$stderr))
stderrsBR <- simplify2array(lapply(resBR, function(x) x$stderr))
stderrsEL <- simplify2array(lapply(resEmp, function(x) x$stderr))

scoresML <- simplify2array(lapply(resML, function(vec) vec$scores["x"]))
scoresBR <- simplify2array(lapply(resBR, function(vec) vec$adjustedScores["x"]))

inadmissibleML <- simplify2array(lapply(resML, function(x) x$inadmissible))
inadmissibleBR <- simplify2array(lapply(resBR, function(x) x$inadmissible))
inadmissibleBC <- simplify2array(lapply(resBC, function(x) x$inadmissible))

convergenceML <- simplify2array(lapply(resML, function(x) x$convergence))
convergenceBR <- simplify2array(lapply(resBR, function(x) x$convergence))
convergenceBC <- simplify2array(lapply(resBC, function(x) x$convergence))

fittedML1 <- simplify2array(lapply(resML, function(x) x$fitted.values))
fittedBR1 <- simplify2array(lapply(resBR, function(x) x$fitted.values))
fittedBC1 <- simplify2array(lapply(resBC, function(x) x$fitted.values))

fittedNames <- paste(rep(seq.int(length.out = length(tots)), each = ncat),
                     seq.int(length.out = ncat), sep = ".")
fittedML <- fittedBR <- fittedBC <- matrix(0, length(fittedNames), ncol(dat))
rownames(fittedML) <- rownames(fittedBR) <- rownames(fittedBC) <- fittedNames
for (i in 1:ncol(dat)) {
  nlevML <- nlevels(resML[[i]]$data$Y)
  namesInML <- as.character(resML[[i]]$data$Y)
  namesInML <- paste(rep(seq.int(length.out = length(tots)), each = nlevML),
                     namesInML, sep = ".")
  nlevBR <- nlevels(resBR[[i]]$data$Y)
  namesInBR <- as.character(resBR[[i]]$data$Y)
  namesInBR <- paste(rep(seq.int(length.out = length(tots)), each = nlevBR),
                     namesInBR, sep = ".")
  fittedML[namesInML, i] <- fittedML1[[i]]
  fittedBC[namesInML, i] <- fittedBC1[[i]]
  fittedBR[namesInBR, i] <- fittedBR1[[i]]
}

boundaryML <- simplify2array(lapply(resML, function(x) x$boundaryEstimates))
boundaryBR <- simplify2array(lapply(resBR, function(x) x$boundaryEstimates))
boundaryBC <- simplify2array(lapply(resBC, function(x) x$boundaryEstimates))


fails <- is.na(scoresBR) | (abs(scoresBR) > 1e-06)
any(fails)

isInf <- (stderrsML > 200) | is.na(stderrsML) | (abs(betasML) > 100)



################################################
### Export the results
################################################
save.image(paste("images/compenu_2x", ncat, "_",
                 paste(tots, collapse = "_"), link, ".RData", sep = ""))


################################################
## Shrinkage plots
################################################
q <- ncat - 1
where <- make.link(link)$linkinv
which <- 1:ncol(dat)
limsx <- c(0, 1)
limsy <- c(0, 1)
low <- -0.1
upp <- 1.1
par(mfrow = c(length(tots), ncat))
for (r in 0:(length(tots)-1)) {
    for (i in 0:q) {
        plot(fittedML[i, 1], fittedBR[i, 1],
             xlab = substitute(hat(pi)[paste(cov,ind)],
             list(ind = i + 1, cov = r + 1)),
             ylab = substitute(tilde(pi)[paste(cov,ind)],
             list(ind = i + 1, cov = r + 1)),
             xlim = limsx, ylim = limsy,
             type = "n")
        title(paste("Category", i + 1))
        abline(0, 1, lty = 2)
        if (i == 0) {
            polygon(x = c(low, low, where(0)), y = c(low, where(0), where(0)),
                    col = "lightgrey", border = NA)
            polygon(x = c(where(0), upp, upp), y = c(where(0), where(0), upp),
                    col = "lightgrey", border = NA)
            abline(v = where(0), lty = 2)
            abline(h = where(0), lty = 2)
        }
        else if (i == q) {
            polygon(x = c(low, low, 1 - where(0)), y = c(low, 1 - where(0), 1 - where(0)),
                    col = "lightgrey", border = NA)
            polygon(x = c(1 - where(0), upp, upp), y = c(1 - where(0), 1 - where(0), upp),
                    col = "lightgrey", border = NA)
            abline(v = 1 - where(0), lty = 2)
            abline(h = 1 - where(0), lty = 2)
        }
        else {
            polygon(x = c(low, upp, upp),
                    y = c(low, low, upp),
                    col = "lightgrey", border = NA)
        }
        points(fittedML[1 + r*ncat + i, which],
               fittedBR[1 + r*ncat + i, which],
               pch = 16, cex = 0.5)
    }
}

